from octostar.client import make_client, UserContext, User
from functools import wraps
from streamlit_octostar_research.desktop import whoami
from streamlit.runtime.scriptrunner import get_script_run_ctx
import hashlib
import streamlit as st
import time
import base64
import json

def impersonating_running_user(**client_kwargs):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            client = as_running_user(**client_kwargs).user.client
            kwargs['client'] = client
            return func(*args, **kwargs)
        return wrapper
    return decorator

def as_running_user(force_refresh=False):
    session = st.session_state.setdefault("__run_user", {})
    script_ctx = get_script_run_ctx()
    current_time = time.time()

    def should_refresh():
        return (
            force_refresh
            or id(script_ctx.script_requests) != session.get('prev_run_id')
            or current_time > session.get('token_expiry', 0) - 300  # 5 minutes buffer
        )

    if not should_refresh() and 'prev_user' in session:
        return UserContext(session['prev_user'])

    running_user = whoami()
    if not running_user:
        return UserContext(session.get('prev_user')) if 'prev_user' in session else st.stop()

    user_hash = int(hashlib.md5(running_user['os_jwt'].encode('utf-8')).hexdigest(), 16)
    if 'prev_user' not in session or hash(session['prev_user']) != user_hash:
        client = make_client(fixed_jwt=running_user['os_jwt'])
        user = User(client)
        session['prev_user'] = user
        session['token_expiry'] = user.jwt_expires_at

    session['prev_run_id'] = id(script_ctx.script_requests)
    return UserContext(session['prev_user'])


import os
import yaml
import json
import base64

from .entities_parser import EntitiesParser, EntitiesParserRuleset
from .parameters import ConsolidationParameters
#from octostar.client import impersonating_launching_user, dev_mode

#@impersonating_launching_user()
#@dev_mode(os.getenv('OS_DEV_MODE'))
def parse_root_nodes_to_linkchart(endpoint, source, body, transformResult, addToGraph,
                                  guess_entries_fn, guess_rulesets_fn, guess_entity_fn, get_parsed_label_fn, process_invalid_fn, ruleset_params, parser_name, client=None):
    # The correct ontological path should be source -> entry (api_response, non-parsed) -> account (root node, parsed) -> emails, people... (non-root nodes, parsed)
    # However, because this looks like a longer path in the linkchart and the requirement is not to while still having a two-steps procedure,
    # we need a way to replace the api_response with the parsed root.
    # Also because the linkchart will bug out by showing two nodes with the same UUID if they have different entity type
    # (the actual id is built with both but no way to externally control this),
    # the only solution (for now) is to compute the parsed nodes immediately to get the entity type and UUID they imply, but still show the raw data.
    # In the future hiding the middle node should really be a styling feature
    entries = guess_entries_fn(endpoint, transformResult)
    records = list()
    relationships = list()
    for entry in entries:
        guessed_entity = guess_entity_fn(endpoint, entry)
        ruleset_files, ruleset_parameters = guess_rulesets_fn(endpoint, entry)
        if isinstance(ruleset_params, list):
            for idx, ruleset_parameter in enumerate(ruleset_parameters):
                if ruleset_parameter and not ruleset_params[idx]:
                    pass
                else:
                    ruleset_parameters[idx] = ruleset_params[idx]
        parser = EntitiesParser([EntitiesParserRuleset(os.path.join("rulesets", ruleset_files[i]),
                                                       ruleset_parameters[i]) for i in range(len(ruleset_files))],
                                client=client)
        new_entry_entities, _ = parser.apply_rules(entry, endpoint, body, guessed_entity)
        new_entry_entities = list(filter(lambda x: not x.parents, new_entry_entities))
        processed_entities, processed_relationships = parser.validate_and_format_entities(new_entry_entities, [])
        processed_entities, processed_relationships = process_invalid_fn(processed_entities, processed_relationships)
        new_linkchart_elements = parser.to_linkchart(processed_entities, processed_relationships, get_parsed_label_fn)
        for new_linkchart_element in new_linkchart_elements['records']:
            records.append({
                "#endpoint": endpoint,
                "#original_entry": base64.b64encode(json.dumps(entry).encode('utf-8')).decode('utf-8'),
                "#parser": parser_name,
                "#guessed_entity": guessed_entity,
                "#is_empty": not bool(entry),
                "#input": body,
                "#ruleset_parameters": ruleset_parameters,
                **new_linkchart_element
            })
            relationships.append({
                'from': source['os_entity_uid'],
                'to': new_linkchart_element['os_entity_uid'],
                'label': endpoint
            })
    if len(records) == 1 and ('#is_empty' not in records[0] or records[0]['#is_empty']):
        records = list()
        relationships = list()
    addToGraph['records'].extend(records)
    addToGraph['relationships'].extend(relationships)
    addToGraph = {'records': addToGraph['records'],
                  'relationships': addToGraph['relationships']}
    return addToGraph

#@impersonating_launching_user()
#@dev_mode(os.getenv('OS_DEV_MODE'))
def parse_all_nodes_to_linkchart(sources, addToGraph, guess_rulesets_fn, get_parsed_label_fn, process_invalid_fn, client=None):
    with open(os.path.join(os.getcwd(), "rulesets", "consolidation_parameters.yaml")) as file:
        consolidation_params_dict = yaml.safe_load(file)
    consolidation_parameters = ConsolidationParameters(equality_tolerances=consolidation_params_dict.get('tolerances', dict()).get('equality', dict()),
                                                       similarity_tolerances=consolidation_params_dict.get('tolerances', dict()).get('similarity', dict()),
                                                       keep_roots=True)
    records = list()
    relationships = list()
    parsed_entities = list()
    parsed_relationships = list()
    for source in sources:
        endpoint = source["#endpoint"]
        guessed_entity = source["#guessed_entity"]
        body = source['#input']
        entry = json.loads(base64.b64decode(source["#original_entry"].encode('utf-8')).decode('utf-8'))
        ruleset_files, _ = guess_rulesets_fn(endpoint, source)
        ruleset_parameters = source["#ruleset_parameters"]
        parser = EntitiesParser([EntitiesParserRuleset(os.path.join("rulesets", ruleset_files[i]),
                                                       ruleset_parameters[i]) for i in range(len(ruleset_files))],
                                client=client)
        new_entities, new_relationships = parser.apply_rules(entry, endpoint, body, guessed_entity)
        parsed_entities.extend(new_entities)
        parsed_relationships.extend(new_relationships)
    parser = EntitiesParser([])
    consolidated_entities, consolidated_relationships, _ = parser.consolidate_entities(parsed_entities, parsed_relationships, consolidation_parameters)
    processed_entities, processed_relationships = parser.validate_and_format_entities(consolidated_entities, consolidated_relationships)
    processed_entities, processed_relationships = process_invalid_fn(processed_entities, processed_relationships)
    new_linkchart_elements = parser.to_linkchart(processed_entities, processed_relationships, get_parsed_label_fn)
    for new_linkchart_element in new_linkchart_elements['records']:
        records.append({
            "#is_empty": not bool(entry),
            **new_linkchart_element
        })
    relationships.extend(new_linkchart_elements['relationships'])
    if len(records) == 1 and ('#is_empty' not in records[0] or records[0]['#is_empty']):
        records = list()
        relationships = list()
    for record in records:
        del record['#is_empty']
    addToGraph['records'].extend(records)
    addToGraph['relationships'].extend(relationships)
    addToGraph = {'records': addToGraph['records'],
                  'relationships': addToGraph['relationships']}
    return addToGraph


from .generics import CombineField
from .info import CombineFieldInfo, EntityInfo, RelationshipInfo
from itertools import islice
import uuid
import json

class CombineFieldLast(CombineField):
    def __init__(self, field, primitive_types_only=True, **params):
        super().__init__(field, **{**params, **{'primitive_types_only': primitive_types_only}})
    
    def combine_field(self, data, entity):
        primitive_types_only = False
        if 'primitive_types_only' in self.params and self.params['primitive_types_only']:
            primitive_types_only = True
        if data:
            last_dict_elem = next(islice(iter(data), len(data)-1, len(data)))
            last_elem = data[last_dict_elem][-1]
            if primitive_types_only and isinstance(last_elem, (dict, list)):
                if len(last_elem) == 1:
                    last_elem = next(last_elem)
                else:
                    raise ValueError("Primitive type only are allowed in output for field " + self.field)
            if last_elem:
                return last_elem
            else:
                return ''
        else:
            return ''

class CombineFieldFirst(CombineField):
    def __init__(self, field, primitive_types_only=True, **params):
        super().__init__(field, **{**params, **{'primitive_types_only': primitive_types_only}})
    
    def combine_field(self, data, entity):
        primitive_types_only = False
        if 'primitive_types_only' in self.params and self.params['primitive_types_only']:
            primitive_types_only = True
        if data:
            first_dict_elem = next(islice(iter(data), 0, 1))
            first_elem = data[first_dict_elem][0]
            if primitive_types_only and isinstance(first_elem, (dict, list)):
                if len(first_elem) == 1:
                    first_elem = next(first_elem)
                else:
                    raise ValueError("Primitive type only are allowed in output for field " + self.field)
            return first_elem
        else:
            return ''

class CombineFieldCustom(CombineField):
    def __init__(self, field, custom_fn, **params):
        super().__init__(field, **params)
        self.custom_fn = custom_fn
    
    def combine_field(self, data, entity):
        if data:
            return self.custom_fn(data, entity)
        else:
            return ''

class CombineFieldJSON(CombineField):
    def __init__(self, field, indent=None, **params):
        super().__init__(field, **params)
        self.params = {**self.params, **{'indent': indent}}
    def combine_field(self, data, entity):
        if data:
            if 'indent' in self.params and self.params['indent']:
                indent = self.params['indent']
            else:
                indent = 4
            return json.dumps(data, indent=indent)
        else:
            return ''

class CombineFieldConcat(CombineField):
    def concat(self, data, dict_sep='', list_sep='', as_flat_list=False):
        def _list_flatten(in_list):
            in_list = list(in_list)
            if in_list == []:
                return in_list
            if isinstance(in_list[0], list):
                return _list_flatten(in_list[0]) + _list_flatten(in_list[1:])
            return in_list[:1] + _list_flatten(in_list[1:])
        if as_flat_list:
            return _list_flatten([[elem for elem in ls] for ls in data.values()])
        else:
            return dict_sep.join([list_sep.join([elem for elem in ls]) for ls in data.values()])
    def __init__(self, field, dict_sep='', list_sep='', as_flat_list=False, **params):
        super().__init__(field, **params)
        self.params = {**self.params, **{'as_flat_list': as_flat_list}}
        self.list_sep = list_sep
        self.dict_sep = dict_sep
    def combine_field(self, data, entity):
        if data:
            if 'as_flat_list' in self.params and self.params['as_flat_list']:
                as_flat_list = True
            else:
                as_flat_list = False
            return self.concat(data, self.dict_sep, self.list_sep, as_flat_list)
        else:
            return ''
        
class CombineFieldSpawn(CombineField):
    def __init__(self, field, inner_combiner=None, **params):
        super().__init__(field, **params)
        self.inner_combiner = inner_combiner
        if not self.inner_combiner:
            self.inner_combiner = CombineFieldCustom(field, lambda d, e: d) # Special combiner that should return a list or dict of lists
    def combine_field(self, data, entity):
        def _list_flatten(in_list):
            in_list = list(in_list)
            if in_list == []:
                return in_list
            if isinstance(in_list[0], list):
                return _list_flatten(in_list[0]) + _list_flatten(in_list[1:])
            return in_list[:1] + _list_flatten(in_list[1:])
        value_dicts = dict()
        all_params = list()
        entity_relationships = ([rel for rel in self.parser.curr_relationships if rel.concept_from == entity.unique_id],
                                [rel for rel in self.parser.curr_relationships if rel.concept_to == entity.unique_id])
        for field_key, field_value in entity.fields.items():
            if isinstance(field_value, CombineFieldInfo) and isinstance(field_value.combiner, CombineFieldSpawn):
                values = field_value.combiner.inner_combiner.combine_field(field_value.data, entity)
                value_dicts[field_key] = dict()
                if isinstance(values, list):
                    value_dicts[field_key]['-1'].extend(values)
                    all_params.append('-1')
                elif isinstance(values, dict):
                    for key, elems in values.items():
                        if key not in value_dicts:
                            value_dicts[field_key][key] = list()
                        value_dicts[field_key][key].extend(elems)
                        all_params.append(key)
        all_params = {param: max([len(inner_dict.get(param, [])) for inner_dict in value_dicts.values()]) for param in all_params}
        value_lists = dict()
        for field_key, inner_dicts in value_dicts.items():
            if field_key not in value_lists:
                value_lists[field_key] = list()
            for param_name in inner_dicts.keys():
                inner_dicts[param_name].extend([None]*(all_params[param_name]-len(inner_dicts[param_name])))
        keys = list(value_dicts.keys())
        value_lists = [_list_flatten([inner_dict[key] for key in sorted(list(inner_dict.keys()))]) for inner_dict in value_dicts.values()]
        if value_lists:
            maxlen = max([len(ls) for ls in value_lists])
            [ls.extend([None]*(maxlen-len(ls))) for ls in value_lists]
            combined_fields = [{keys[i]: value_lists[i][j] for i in range(len(keys))} for j in range(maxlen)]
            first_elem_fields = combined_fields.pop(0)
            entity.fields = {**entity.fields, **first_elem_fields}
            for combined_fields_elem in combined_fields:
                new_fields = entity.fields
                new_fields = {**entity.fields, **combined_fields_elem}
                new_entity = EntityInfo(str(uuid.uuid4()), entity.concept_name, new_fields, parents=entity.parents)
                new_rels = [RelationshipInfo(new_entity.unique_id, rel.concept_to, rel.name) for rel in entity_relationships[0]] +\
                        [RelationshipInfo(rel.concept_from, new_entity.unique_id, rel.name) for rel in entity_relationships[1]]
                self.parser.curr_entities.append(new_entity)
                self.parser.curr_relationships.extend(new_rels)



import asyncio
from fastapi import Request
from fastapi.responses import JSONResponse, StreamingResponse
from pydantic import BaseModel
from typing import Union, List, Optional, Literal
from io import BytesIO
import zipfile
import json
from pydantic import ValidationError
from fastapi.exceptions import RequestValidationError
from starlette.exceptions import HTTPException as StarletteHTTPException
import ast
import uuid
from functools import wraps, partial
import traceback
from copy import copy

MAX_ERROR_MESSAGE_BYTES = 256
MAX_ERROR_TRACEBACK_BYTES = 10240
DEFAULT_PROCESSOR_SUFFIX = 'main'


class CommonParsers(object):
    async def parse_form_list(form_data: str,
                              format=Literal['csv', 'python']) -> Optional[List[str]]:
        try:
            if not form_data:
                return None
            if format == 'csv':
                return form_data.split(',')
            elif format == 'python':
                return ast.literal_eval(form_data)
        except:
            raise ValidationError("Parsing failed in parse_from_list!")

class Route(object):
    def route(route, **route_kwargs):
        def decorator(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                return await func(*args, **kwargs)
            func_name = func.__name__
            wrapper._route_kwargs = route_kwargs
            wrapper._route_funcname = func_name
            if getattr(route, func_name, None):
                raise ValueError("Cannot have two FastAPI routes with the same name within the same Route!")
            setattr(route, func_name, wrapper)
            route.routed_funcs.append(wrapper)
            return wrapper
        return decorator

    def __init__(self, app, router=None):
        self.app = app
        self._router = router
        self.routed_funcs = []
        
    def define_routes(self):
        raise NotImplementedError()
    
    @property
    def router(self):
        return self._router or self.app
    
    def include_routes(self, router_mapping=None):
        for func in self.routed_funcs:
            route_kwargs = copy(func._route_kwargs)
            router = self.router
            if route_kwargs.get('router'):
                router = route_kwargs['router']
            elif router_mapping:
                router = router_mapping.get(func._route_funcname, self.router)
            else:
                router = self.router
            route_kwargs.pop('router', None)
            router.add_api_route(
                **route_kwargs,
                endpoint=func
            )
        return self
        
class CommonModels(object):
    class OKResponseModel(BaseModel):
        message: str = "OK"
        status: str = "success"

    class DataResponseModel(BaseModel):
        data: Union[list, dict]
        status: str = "success"

    class ZipResponseModel(object):
        class Element(object):
            def __init__(self, file: bytes, path: str, filename: str, json: dict):
                self.file = file
                self.path = path
                self.filename = filename
                self.json = json

        def __init__(self, elements: List[Element], paths: Optional[List[str]] = None, filename: Optional[str] = None, status: str = "success"):
            self.elements = elements
            self.paths = paths
            self.status = status
            self.filename = filename

        def get(self):
            zip_buffer = BytesIO()
            with zipfile.ZipFile(zip_buffer, 'w', zipfile.ZIP_DEFLATED) as zip_file:
                for element in self.elements:
                    file_name = f"{element.path}/{element.filename}"
                    zip_file.writestr(file_name, element.file)
                    json_name = f"{element.path}/{element.filename}.dict"
                    zip_file.writestr(json_name, json.dumps(element.json, indent=4))
                for path in self.paths:
                    zip_file.writestr(path + "/", '')
                zip_file.writestr('response.dict', json.dumps({'status': self.status}, indent=4))
            zip_buffer.seek(0)
            headers = {}
            headers['Content-Disposition'] = 'attachment;'
            filename = self.filename or (str(uuid.uuid4()) + ".zip")
            headers['Content-Disposition'] += f' filename="{filename}"'
            return StreamingResponse(zip_buffer, media_type='application/zip', headers=headers)


class DefaultErrorRoute(Route):
    error_responses = {
        500: {
            "description": "Generic Server Error",
            "content": {
                "application/json": {
                    "example": {"message": "Something unexpected went wrong!", "status": "error"}
                }
            }
        },
        501: {
            "description": "Not Implemented",
            "content": {
                "application/json": {
                    "example": {"message": "This method (with these parameters) is not implemented!", "status": "error"}
                }
            }
        },
        400: {
            "description": "Bad Request",
            "content": {
                "application/json": {
                    "example": {"message": "These two lists must have the same length!", "status": "error"}
                }
            }
        },
        422: {
            "description": "Validation Error",
            "content": {
                "application/json": {
                    "example": {"message": "Validation error on 0 -> filename!", "status": "error"}
                }
            }
        },
    }

    def format_error(exc, body=bytes(), debug=False):
        '''Generic Error Handler'''
        status_code = 500
        if isinstance(exc, StarletteHTTPException):
            status_code = exc.status_code
        elif isinstance(exc, NotImplementedError):
            status_code = 501
        elif isinstance(exc, AssertionError):
            status_code = 400
        elif isinstance(exc, ValidationError) or isinstance(exc, RequestValidationError):
            status_code = 422
        try:
            message = exc.message
        except:
            message = str(exc)
        if debug:
            message += "\n" + str(body)
        if len(message) > MAX_ERROR_MESSAGE_BYTES:
            message = message[-MAX_ERROR_MESSAGE_BYTES:]
        tcbk = traceback.format_exception(exc)
        if len(tcbk) > MAX_ERROR_TRACEBACK_BYTES:
            tcbk = tcbk[-MAX_ERROR_TRACEBACK_BYTES:]
        response_content = {"message": message, 'status': 'error'}
        if debug:
            response_content['traceback'] = tcbk
        return JSONResponse(
            status_code=status_code,
            content=response_content
        )

    async def handle_error(request: Request, exc: Exception, debug: bool):
        try:
            body = await request.body()
        except:
            body = bytes()
        return DefaultErrorRoute.format_error(exc, body, debug=debug)

    def add_default_exceptions_handler(fs_app, debug=False):
        async def async_handle_error(request: Request, exc: Exception):
            return await DefaultErrorRoute.handle_error(request, exc, debug)

        fs_app.add_exception_handler(RequestValidationError, async_handle_error)
        fs_app.add_exception_handler(StarletteHTTPException, async_handle_error)
        fs_app.add_exception_handler(Exception, async_handle_error)


class RequestCancelledMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return
        queue = asyncio.Queue()

        async def message_poller(sentinel, handler_task):
            nonlocal queue
            while True:
                message = await receive()
                if message["type"] == "http.disconnect":
                    handler_task.cancel()
                    return sentinel
                await queue.put(message)
        sentinel = object()
        handler_task = asyncio.create_task(self.app(scope, queue.get, send))
        asyncio.create_task(message_poller(sentinel, handler_task))
        try:
            return await handler_task
        except asyncio.CancelledError:
            print("Cancelling request due to disconnect")